from .otBase import BaseTTXConverter


class table_B_A_S_E_(BaseTTXConverter):
    """Baseline table

    The ``BASE`` table contains information needed to align glyphs in
    different scripts, from different fonts, or at different sizes
    within the same line of text.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/base
    """

    pass
